// ======================================================================
//
// StringFile.cpp
// Jason Hughes
//
// copyright 1998 origin systems inc
//
// ======================================================================

#include "first.h"
#include "StringFile.h"

#include "Iff.h"
#include "VString.h"

//--------------------------------------------------

const int StringFileBufferSize = 16000;   // this is the longest string that is allowed to load without crashing

StringFile::StringBattery *StringFile::lpList=NULL;
int                        StringFile::languageTableCount;
StringFile::LanguageInfo   StringFile::langTable[] = {
			{ "English", 1 },
			{ "Spanish", 1 },
			{ "German",  1 },
			{ "French",  1 },
			{ "Japanese",2 },
			{ "Italian", 1 },
			{ "", 0}};  // this signifies the end of the table

//--------------------------------------------------

void StringFile::install(void)
{
	lpList=NULL;
	for (languageTableCount=0; langTable[languageTableCount].charsize; languageTableCount++)  // find the last entry
		;
}

void StringFile::remove(void)
{
	StringBattery *temp;

	DEBUG_FATAL(lpList,("StringFile::remove - partition is STILL ALLOCATED with refCount=%d for partition [%s].\n",lpList->refCount,lpList->filename));
	while (lpList)
	{
		temp=lpList;
		lpList=lpList->next;
		delete temp;
	}
}

StringFile::StringFile(const char *fn)
{
	StringBattery *walker;

	walker=lpList;
	while (walker && stricmp(fn,walker->filename))
		walker=walker->next;

	if (walker)               // found the filename
	{
		data=walker;          // simply assign our data pointer to point to the battery we want to use
		data->refCount++;	  // keep the reference count up to date
		data->access++;       // make sure it knows the number of accesses it's getting for performance reasons
	}
	else                      // haven't loaded the filename yet
	{
		int version;
		Iff iff(fn);

		iff.seekForm(TAG(L,A,N,G));
			if (!iff.seekChunk(TAG(V,E,R,S),Iff::FlagOptional))
				Fatal("StringFile::StringFile() - %s no version chunk.",fn);
			version=iff.read_u_int32();
			iff.exitChunk();

//-----------------------------------
			data=NEW StringBattery;  // create a new string battery and hook it into the static list for the loaders

			data->next=lpList;
			lpList=data;          // link it into the head of the static list

			data->filename=strdup(fn);
			data->refCount=1;
			data->access=1;
//-----------------------------------

			switch (version)
			{
				case 1:
					load_0001(&iff);
					break;
				default:
					Fatal("StringFile::StringFile - unrecognized version %d.",version);
					break;
			}
		iff.exitForm();
	}
}

StringFile::~StringFile(void)
{
	data->refCount--;
	if (!data->refCount)
	{
		delete []data->filename;
		data->filename=NULL;

		for (int loop=0; loop<data->numStrings; loop++)
		{
			delete data->strings[loop];
			data->strings[loop]=NULL;
		}

		delete []data->strings;
		data->strings=NULL;

		delete []data->uniqueIds;
		data->uniqueIds=NULL;

//----------------- clean up static list

		if (lpList==data)     // deleting the head of the list
		{
			lpList=lpList->next;
			delete data;
		}
		else                  // find, then wipe out structure in list
		{
			StringBattery *walker;

			walker=lpList;
			while (walker->next!=data)
				walker=walker->next;
			if (walker->next!=data)
				Fatal("StringBattery not found in ~StringFile.\n");
			walker->next=data->next;
			delete data;
		}
	}
}

void StringFile::load_0001(Iff *iff)
{
	int tmpchrsz;
	void *strBuffer=static_cast<void *>(NEW char[StringFileBufferSize]);

	iff->seekChunk(TAG(H,E,A,D));
	data->maxUniqueId=iff->read_u_int32();
	data->languageId=iff->read_u_int32();
	tmpchrsz=iff->read_u_int32();

	if (data->languageId<languageTableCount)
	{
		DEBUG_FATAL (tmpchrsz != langTable[data->languageId].charsize,("String file does not have correct charsize: %s should be %d instead of %d.",iff->getFilename(),langTable[data->languageId].charsize,tmpchrsz));
		DEBUG_FATAL (tmpchrsz > String::queryCharsize(),("String file does not have correct charsize: %s should be %d instead of %d.",iff->getFilename(),String::queryCharsize(),tmpchrsz));
	}
	else
	{
		DEBUG_FATAL (tmpchrsz > String::queryCharsize(),("String file does not have correct charsize: %s should be %d instead of %d.",iff->getFilename(),String::queryCharsize(),tmpchrsz));
	}

	iff->exitChunk();

	iff->seekChunk(TAG(S,T,R,S));
		for (data->numStrings=0; !iff->endOfChunk(); data->numStrings++)  // fake load and count all strings in this chunk
		{
			iff->read_u_int32();
			iff->read_u_int32();
			switch (tmpchrsz)
			{
				case 1:
					iff->read_string(static_cast<char *>(strBuffer),StringFileBufferSize);
					break;
				case 2:
					iff->read_string16(static_cast<int16 *>(strBuffer),StringFileBufferSize);
					break;
				case 4:
					iff->read_string32(static_cast<int32 *>(strBuffer),StringFileBufferSize);
					break;
				default:
					Fatal("Impossible character size in file %s.",iff->getFilename());
			}
		}

		iff->goToTopOfChunk();  // now we really load stuff after allocating array space

		data->uniqueIds=NEW int[data->numStrings];         // allocate an array for unique IDs
		data->strings=NEW String *[data->numStrings];      // allocate an array of pointers to strings, then really load strings
		for (int loop=0; loop<data->numStrings; loop++)
		{
			iff->read_u_int32();  // trash the timestamp
			data->uniqueIds[loop]=iff->read_u_int32();
			switch (tmpchrsz)
			{
				case 1:
					iff->read_string(static_cast<char *>(strBuffer),StringFileBufferSize);
					break;
				case 2:
					iff->read_string16(static_cast<int16 *>(strBuffer),StringFileBufferSize);
					break;
				case 4:
					iff->read_string32(static_cast<int32 *>(strBuffer),StringFileBufferSize);
					break;
			}
			data->strings[loop]=NEW String(strBuffer,tmpchrsz);  // allocate exactly the right amount of space for this string
		}

		DEBUG_FATAL(iff->chunkLeft(), ("StringFile:: specified string count too small in %s", iff->getFilename()));
	iff->exitChunk();

	delete []static_cast<char *>(strBuffer);
}

//-----------------------------------------------------

const String &StringFile::getStringById(int id) const
{
	DEBUG_FATAL(id<0 || id>=data->numStrings,("StringFile::getString - accessing out of bounds: %d/%d.\n",id,data->numStrings));
	for (int loop=0; loop<data->numStrings; loop++)
		if (data->uniqueIds[loop]==id)
			return *data->strings[loop];

	Fatal("Did not find string by id.  Id#%d.",id);
	return *data->strings[0];  // dummy line
}

const String &StringFile::getStringByIndex(int index) const
{
	DEBUG_FATAL(index<0 || index>=data->numStrings,("StringFile::getString - accessing out of bounds: %d/%d.\n",index,data->numStrings));
	return *data->strings[index];
}
